(*====================================================================
    File rbdeutils.pas

    Components library for Delphi and BCB++.
    , ,   
         BDE.
    Procedures, functions, constants and types
    for working with database (BDE).

    Copyright  1998-2008, PSTG
    http://www.polesoft.ru
    mailto:support@polesoft.ru

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

======================================================================
$Id: rBDEUtils.pas,v 1.7 2008/10/06 11:47:15 khmelevskiy Exp $
======================================================================*)
unit rBDEUtils;

{$I POLARIS.INC}

interface

uses rDBConst, BDE, DBTables, DB, SysUtils, rUtils, rErrorUtils, Classes;

const
{$IFDEF POLARIS_D4}
  DBMasks: array[TTableType] of string[5] = ('', '*.DB', '*.DBF', '*.TXT', '*.DBF');
{$ELSE}
  DBMasks: array[TTableType] of string[5] = ('', '*.DB', '*.DBF', '*.TXT');
{$ENDIF}

//  SQL 
// Execute SQL
procedure ExecuteSQL(ADatabase: TDatabase; SQL: string);

//  /   
// System date&time on DB server
function DBSysDate(ADatabase: TDatabase; WithTime: Boolean): TDateTime;

{$IFNDEF POLARIS_D4}
//   
// Assign parameter (TParam) value
procedure AssignParam(Param: TParam; Value: Variant);

//    
// Clear all TParam's values
procedure ClearParams(Params: TParams);

//     
// Assign parameter (TParam) value from field
procedure AssignParamFromField(Param: TParam; Field: TField);
{$ENDIF}

//     
// Execute stored procedure within transaction
procedure ExecStoredProc(StoredProc: TStoredProc; DB: TDatabase);

//  
// User name
function GetUser(DB: TDatabase): string;

//   BDE
// BDE errors handler
procedure BDEErrorEHProc(E: Exception; Handler: TrError);

//  
// Macros substitution
function FormatDBCode(ADataBase: TDataBase; AText, AUserName: string; APrefix: Char = '%'): string;

implementation

procedure ExecuteSQL(ADatabase: TDatabase; SQL: string);
var
  FStmtHandle: hDBIStmt;
begin
  Check(DbiQAlloc(ADataBase.Handle, qrylangSQL, FStmtHandle));
  try
    Check(DbiQPrepare(FStmtHandle, {$IFDEF POLARIS_D12}PAnsiChar{$ELSE}PChar{$ENDIF}(SQL)));
  except
    DbiQFree(FStmtHandle);
    FStmtHandle := nil;
    raise;
  end;
  Check(DbiQExec(FStmtHandle, nil));
  Check(DbiQFree(FStmtHandle));
end;

function DBSysDate(ADatabase: TDatabase; WithTime: Boolean): TDateTime;
var
  driver: string;
begin
  if ADatabase.AliasName <> '' then
    driver := ADatabase.Session.GetAliasDriverName(ADatabase.AliasName)
  else
    driver := ADatabase.DriverName;
  if driver = 'INTRBASE' then
    with TQuery.Create(nil) do
      try
        DatabaseName := ADatabase.DatabaseName;
        SQL.Text := srSQLIBSysDate;
        Open;
        Result := Fields[0].AsDateTime;
        if not WithTime then
          Result := StrToDate(DateToStr(Result));
        Close;
      finally
        Free;
      end
  else
    if WithTime then Result := SysUtils.Now else Result := SysUtils.Date;
end;

procedure ExecStoredProc(StoredProc: TStoredProc; DB: TDatabase);
var
  b: Boolean;
begin
  with StoredProc, DB do begin
    b := InTransaction;
    if not b then StartTransaction;
    try
      ExecProc;
      if not b then Commit;
    except
      if not b then Rollback;
      raise;
    end;
  end;
end;

function GetUser(DB: TDatabase): string;
begin
  if Assigned(DB) and DB.Connected then
    Result := DB.Params.Values[szUSERNAME]+' '+srDBUserName
  else
    Result := GetOSUser+' '+srOSUserName;
end;

{$IFNDEF POLARIS_D4}
procedure AssignParam(Param: TParam; Value: Variant);
begin
  if VarIsNull(Value) then Param.Clear else Param.Value := Value;
  Param.Bound := True;
end;

procedure ClearParams(Params: TParams);
var
  I: Integer;
begin
  for I := 0 to Params.Count - 1 do Params[I].Clear;
end;

procedure AssignParamFromField(Param: TParam; Field: TField);
begin
  if Field.IsNull then Param.Clear else Param.Value := Field.Value;
  Param.Bound := True;
end;
{$ENDIF}

procedure BDEErrorEHProc(E: Exception; Handler: TrError);
var
  i: Integer;
  mes: string;
begin
  with Handler do begin
    PreMessage := srBDEErrorTitle;
    with E as EDBEngineError do begin
      for i:=0 to ErrorCount-1 do begin  { loop by errors }
        mes := Errors[i].Message;
        // delete last LF or CR in message
        while (Length(mes) > 0) and (Copy(mes, Length(mes), 1)[1] in [#10,#13]) do
          System.Delete(mes, Length(mes), 1);
        if Length(mes) = 0 then Continue;
        if (Handler.Message = '') and Assigned(FIniFile) and
           ((i = ErrorCount-1) or (Errors[i].NativeError <> 0)) then begin
          Handler.Message := ConvertMessage(Errors[i].Message);
          if Handler.Message = '' then Handler.Message := Errors[i].Message;
        end;
        with Errors[i] do
          if NativeError = 0 then
            Description.Add(Format(srBDEErrorDesc,
                                    [ErrorCode, Category, SubCode,mes]))
          else
            Description.Add(Format(srSQLErrorDesc, [NativeError, mes]));
      end;
      if Handler.Message = '' then begin
        Handler.Message := Errors[ErrorCount-1].Message;
        SaveLog := True;
      end;
      // list of active databases
      mes := '';
      for i:=0 to Session.DatabaseCount-1 do
        if Session.Databases[i].Connected then
          if Session.Databases[i].AliasName = '' then
            mes := mes + Session.Databases[i].DatabaseName + ';'
          else
            mes := mes + Session.Databases[i].AliasName + ';';
      if Length(mes) > 0 then
        Description.Add(Format(srActiveDB, [Copy(mes, 1, Length(mes)-1)]));
    end;
  end;
end;

procedure OLEDBErrorEHProc(E: Exception; Handler: TrError);
var
  i: Integer;
begin
  with Handler do begin
    PreMessage := srBDEErrorTitle;
    i := Pos('%s', srBDEGeneralSQLError) - 1;
    if (Copy(srBDEGeneralSQLError, 1, i) = Copy(E.Message, 1, i)) then
      Message := ConvertMessage(Copy(E.Message, i + 3, Length(E.Message)))
    else
      Message := ConvertMessage(E.Message);
    if Message = '' then Message := E.Message
    else
      Description.Text := E.Message;
  end;
end;

function FormatDBCode(ADataBase: TDataBase; AText, AUserName: string; APrefix: Char = '%'): string;
var
  SysDate: TDateTime;
  SValue: string;
  i, n, j: Integer;
  paramList: TStringList;
const
  MaxFuncs = 11;
  TFuncs: array[0..MaxFuncs] of string =
    ('yyyy', 'yy', 'mm', 'dd', 'hh', 'nn', 'ss', 'compname', 'dbusername',
     'dbuser', 'dbname', 'osuser');
begin
  Result := AText;
  SysDate := DBSysDate(ADataBase, True);
  for i := 0 to MaxFuncs do begin
    while True do begin
      n := Pos(APrefix + AnsiLowerCase(TFuncs[i]), AnsiLowerCase(Result));
      if n = 0 then Break;
      case i of
        0..6: SValue := FormatDateTime(TFuncs[i], SysDate);
        7: SValue := GetCompName;
        8: SValue := AUserName;
        9: SValue := ADataBase.Params.Values[szUSERNAME];
       10: begin
             paramList := TStringList.Create;
             try
               Session.GetAliasParams(ADataBase.DatabaseName, paramList);
               SValue := paramList.Values[szSERVERNAME];
               for j := Length(SValue) downto 0 do
                 if (SValue[j] = '/') or (SValue[j] = '\') then Break;
               SValue := ChangeFileExt(Copy(SValue, j + 1, MaxInt), '');
             finally
               paramList.Free;
             end;
           end;
       11: SValue := GetOSUser;
      end;
      Result := Stuff(Result, n, Length(TFuncs[i]) + 1, SValue);
    end;
  end;
end;

initialization
  AddError('EDBEngineError', BDEErrorEHProc);
  AddOrReplaceError('EOLEException', OLEDBErrorEHProc);
end.

