(*====================================================================
    File rClasses.pas

    Components library for Delphi, Kylix and BCB++.
    .
    Classes.

    Copyright  1998-2008, PSTG
    http://www.polesoft.ru
    mailto:support@polesoft.ru

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

======================================================================
$Id: rClasses.pas,v 1.8 2008/09/05 09:00:45 savchenko Exp $
======================================================================*)
unit rClasses;

{$I POLARIS.INC}

interface

uses
{$ifdef mswindows}
  Windows,
{$endif}
  SysUtils, Classes, Math, rUtils, rConst,
  {$IFDEF POLARIS_D6}RxStrUtils, Variants{$ELSE}StrUtils{$ENDIF};

const
  crListDelim = ',';

type

  { TVariantList }

  TVarListAction = (vlnAdd, vlnInsert, vlnExtract, vlnDelete, vlnChange, vlnClear);

  TVariantList = class(TObject)
  private
    FList: TList;
    function Get(const Index: Integer): Variant;
    procedure Put(const Index: Integer; const Value: Variant);
    function GetCount: Integer;
  protected
    procedure Notify(Value: Variant; Action: TVarListAction); virtual;
  public
    constructor Create;
    destructor Destroy; override;
    function IndexOf(Value: Variant): Integer;
    function Add(Item: Variant): Integer; virtual;
    function Remove(Item: Variant): Integer; virtual;
    function Extract(Item: Variant): Variant; virtual;
    function First: Variant;
    function Last: Variant;
    procedure Clear; virtual;
    procedure Toggle(Item: Variant);
    procedure Delete(Index: Integer); virtual;
    procedure Insert(Index : Integer; Item: Variant); virtual;
    procedure LoadFromString(S: string; Delims: TCharSet {$IFDEF POLARIS_D4}= [crListDelim]{$ENDIF};
      const AVarType: TVarType {$IFDEF POLARIS_D4}= varInteger{$ENDIF});
    function SaveToString(Delim: string {$IFDEF POLARIS_D4}= crListDelim{$ENDIF}): string;
    procedure Assign(Source: TVariantList);
    property Count: Integer read GetCount;
    property Items[const Index: Integer]: Variant read Get write Put; default;
  end;

{$IFDEF POLARIS_D5}
  TrListNotifyEvent = procedure
    (Sender: TObject; Ptr: Pointer; Action: TListNotification) of object;
{$ENDIF}

  TrList = class(TList)
{$IFDEF POLARIS_D5}
  private
    FOnNotify: TrListNotifyEvent;
  protected
    procedure SetOnNotify(const aValue: TrListNotifyEvent); virtual;
    procedure Notify(Ptr: Pointer; Action: TListNotification); override;
  public
    property OnNotify: TrListNotifyEvent read FOnNotify write SetOnNotify;
  end
{$ENDIF};

  TrNodeClass = class of TrTree;

  TrTree = class(TrList)
  private
    FNodes: TrList;
    FNodeData: Pointer;
    FNodeClass: TrNodeClass;
    FParentNode: TrTree;
    function GetNodeObject(const aIndex: Integer): TrTree;
    function GetNodesCount: Integer;
    procedure __addnode(const aItem: Pointer; out aIndex; out aNode);
{$IFDEF POLARIS_D5}
    procedure __notifynode(Sender: TObject; Ptr: Pointer; Action: TListNotification);
{$ENDIF}
  protected
{$IFDEF POLARIS_D5}
    procedure SetOnNotify(const aValue: TrListNotifyEvent); override;
{$ENDIF}
    function GetNodeData(const aIndex: Integer): Pointer;
    procedure PutNodeData(const aIndex: Integer; const aItem: Pointer);
  public
    procedure AfterConstruction; override;
    procedure BeforeDestruction; override;
    //
    function AddNode(const aItem: Pointer): Integer;
    function CheckNode(const aItem: Pointer): TrTree;
    function CheckOrAddNode(const aItem: Pointer): TrTree;
    procedure ExchangeNodes(const aIndex1, aIndex2: Integer);
    function IndexOfNode(const aItem: Pointer): Integer;
    function NodeExists(const aItem: Pointer): Boolean;
    procedure DeleteNode(const aIndex: Integer);
    function ItemExists(const aItem: Pointer): Boolean;
    procedure ClearItems;
    procedure ClearNodes;
    procedure Clear; override;
    //
    property NodeData: Pointer read FNodeData write FNodeData;
    property Nodes[const aIndex: Integer]: Pointer read GetNodeData write PutNodeData;
    property NodesCount: Integer read GetNodesCount;
    property NodeObjects[const aIndex: Integer]: TrTree read GetNodeObject;
    property NodeClass: TrNodeClass read FNodeClass write FNodeClass;
    property ParentNode: TrTree read FParentNode;
  end;

  TrLogFile = class
  private
    FStream: TFileStream;
    FCloseImmediately: Boolean;
    FLogFileName: string;
    procedure SetLogFileName(const Value: string);
    procedure SetCloseImmediately(const Value: Boolean);
  public
    procedure AfterConstruction(); override;
    procedure WriteLog(const aText: string; AppendTime: Boolean = True);
    procedure Close();
    procedure BeforeDestruction(); override;
    //
    property LogFileName: string read FLogFileName write SetLogFileName;
    property CloseImmediately: Boolean read FCloseImmediately write SetCloseImmediately;
  end;

  // optimized string expansion
  IAppendString = interface
  ['{CDD91CD9-FF58-4D30-BDEA-4BF7EA2D001D}']
    function GetAsString: string;
    function GetCapacity: Integer;
    procedure SetCapacity(const Value: Integer);
    procedure BeginAppend();
    procedure Append(const aString: string);
    procedure EndAppend();
    property AsString: string read GetAsString;
    property Capacity: Integer read GetCapacity write SetCapacity;
  end;

  TAppendString = class(TInterfacedObject, IAppendString)
  private
    FString: string;
    FLength: Integer;
    FAppendFlag: Boolean;
    function GetAsString: string;
    function GetCapacity: Integer;
    procedure SetCapacity(const Value: Integer);
  public
    class function CreateNew(): IAppendString;
    //
    procedure BeginAppend();
    procedure Append(const aString: string);
    procedure EndAppend();
    //
    property AsString: string read GetAsString;
    property Capacity: Integer read GetCapacity write SetCapacity;
  end;


implementation

{ TVariantList }

constructor TVariantList.Create;
begin
  inherited Create;
  FList := TList.Create;
end;

destructor TVariantList.Destroy;
var
  i: Integer;
begin
  for i := 0 to FList.Count - 1 do Dispose(PVariant(FList.Items[I]));
  FList.Free;
  FList := nil;
  inherited Destroy;
end;

function TVariantList.Add(Item: Variant): Integer;
var
  P : PVariant;
begin
  //if IndexOf(Item) < 0 then begin
    New(P);
    VarCopy(P^, Item);
    Result := FList.Add(P);
  {end
  else Result := -1;}
  Notify(Item, vlnAdd);
end;

procedure TVariantList.Clear;
var
  I: Integer;
begin
  for i := 0 to FList.Count - 1 do Dispose(PVariant(FList.Items[I]));
  FList.Clear;
  FList.Pack;
  Notify(Null, vlnClear);
end;

procedure TVariantList.Delete(Index: Integer);
var
  P: PVariant;
begin
  P := PVariant(FList.Items[Index]);
  FList.Delete(Index);
  FList.Pack;
  Notify(P^, vlnDelete);
  Dispose(P);
end;

procedure TVariantList.Put(const Index: Integer; const Value: Variant);
begin
  if (Index >= 0) and (Index < FList.Count) then
    VarCopy(PVariant(FList.Items[Index])^, Value);
  Notify(Value, vlnChange);
end;

function TVariantList.Get(const Index: Integer): Variant;
begin
  Result := Null;
  if (Index >= 0) and (Index < FList.Count) then
    Result := PVariant(FList.Items[Index])^;
end;

function TVariantList.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TVariantList.IndexOf(Value: Variant): Integer;
begin
  Result := 0;
  while (Result < FList.Count) and (CompareVar(Items[Result], Value) <> 0) do
    Inc(Result);
  if Result = FList.Count then Result := -1;
end;

procedure TVariantList.Notify(Value: Variant; Action: TVarListAction);
begin
end;

function TVariantList.Remove(Item: Variant): Integer;
begin
  Result := IndexOf(Item);
  if Result >= 0 then
    Delete(Result);
end;

procedure TVariantList.Toggle(Item: Variant);
var
  Index: Integer;
begin
  Index := IndexOf(Item);
  if Index < 0 then Add(Item)
  else Delete(Index);
end;

function TVariantList.Extract(Item: Variant): Variant;
var
  I: Integer;
begin
  Result := Null;
  I := IndexOf(Item);
  if I >= 0 then
  begin
    Result := Item;
    Delete(I);
    Notify(Result, vlnExtract);
  end;
end;

function TVariantList.First: Variant;
begin
  Result := Get(0);
end;

function TVariantList.Last: Variant;
begin
  Result := Get(Count - 1);
end;

procedure TVariantList.Insert(Index : Integer; Item: Variant);
var
  P: PVariant;
begin
  New(P);
  VarCopy(P^, Item);
  FList.Insert(Index, P);
  Notify(Item, vlnInsert);
end;

procedure TVariantList.LoadFromString(S: string; Delims: TCharSet;
  const AVarType: TVarType);
var
  I: Integer;
  Value: Variant;
begin
  S := DelChars(S, '''');
  for i := 0 to WordCount(S, Delims) - 1 do begin
    case AVarType of
      varInteger, varSmallint: Value := StrToInt(ExtractWord(i + 1, S, Delims));
      varInt64: Value := StrToInt64(ExtractWord(i + 1, S, Delims));
      varDate: Value := StrToDateTime(ExtractWord(i + 1, S, Delims));
      varSingle: Value := StrToFloat(ExtractWord(i + 1, S, Delims));
      varDouble: Value := StrToFloat(ExtractWord(i + 1, S, Delims));
      varCurrency: Value := StrToFloat(ExtractWord(i + 1, S, Delims));
      else Value := ExtractWord(i + 1, S, Delims);
    end;
    Add(Value);
  end;
end;

function TVariantList.SaveToString(Delim: string): string;
var
  J: Integer;
begin
  Result := '';
  for J := 0 to Count - 1 do
    if VarType(Items[J]) = varDate then
      Result := Result + '''' + DateToStr(Items[J]) + '''' + Delim
    else
    if VarType(Items[J]) = varString then
      Result := Result + '''' + string(Items[J]) + '''' + Delim
    else
      Result := Result + string(Items[J]) + Delim;
  System.Delete(Result, Length(Result), 1);
end;

procedure TVariantList.Assign(Source: TVariantList);
var
  i: Integer;
begin
  if not Assigned(Source) then Exit;
  Clear;
  for i := 0 to Source.Count - 1 do
    Add(Source[i]);
end;

{ TrList }
{$IFDEF POLARIS_D5}
procedure TrList.SetOnNotify(const aValue: TrListNotifyEvent);
begin
  FOnNotify := aValue;
end;

procedure TrList.Notify(Ptr: Pointer; Action: TListNotification);
begin
  inherited;
  if Assigned(OnNotify) then OnNotify(Self, Ptr, Action);
end;
{$ENDIF}

{ TrTree }

procedure TrTree.AfterConstruction;
begin
  inherited;
  FNodeClass := TrTree;
  FNodes := TrList.Create;
{$IFDEF POLARIS_D5}
  FNodes.OnNotify := __notifynode;
{$ENDIF}
end;

{$IFDEF POLARIS_D5}
procedure TrTree.__notifynode(Sender: TObject; Ptr: Pointer; Action: TListNotification);
begin
  if Assigned(FOnNotify) then FOnNotify(Self, TrTree(Ptr).FNodeData, Action);
  if Action = lnDeleted then TrTree(Ptr).Free;
end;
{$ENDIF}

procedure TrTree.__addnode(const aItem: Pointer; out aIndex; out aNode);
var
  n: TrTree;
  i: Integer;
begin
  n := FNodeClass.Create;
  n.FNodeData := aItem;
{$IFDEF POLARIS_D5}
  n.OnNotify := OnNotify;
{$ENDIF}
  n.NodeClass := FNodeClass;
  n.FParentNode := Self;
  i := FNodes.Add(n);
  if Assigned(@aIndex) then Integer(aIndex) := i;
  if Assigned(@aNode) then TrTree(aNode) := n;
end;

function TrTree.AddNode(const aItem: Pointer): Integer;
begin
  __addnode(aItem, Result, nil^);
end;

function TrTree.GetNodesCount: Integer;
begin
  Result := FNodes.Count;
end;

function TrTree.GetNodeData(const aIndex: Integer): Pointer;
begin
  Result := NodeObjects[aIndex].FNodeData;
end;

procedure TrTree.PutNodeData(const aIndex: Integer; const aItem: Pointer);
begin
  NodeObjects[aIndex].FNodeData := aItem;
end;

procedure TrTree.ExchangeNodes(const aIndex1, aIndex2: Integer);
begin
  FNodes.Exchange(aIndex1, aIndex2);
end;

function TrTree.IndexOfNode(const aItem: Pointer): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i := 0 to FNodes.Count - 1 do
    if Nodes[i] = aItem then begin
      Result := i;
      Break;
    end;
end;

function TrTree.CheckNode(const aItem: Pointer): TrTree;
var
  i: Integer;
begin
  for i := 0 to FNodes.Count - 1 do begin
    Result := NodeObjects[i];
    if Result.FNodeData = aItem then Exit;
  end;
  Result := nil;
end;

function TrTree.CheckOrAddNode(const aItem: Pointer): TrTree;
begin
  Result := CheckNode(aItem);
  if not Assigned(Result) then
    __addnode(aItem, nil^, Result);
end;

function TrTree.NodeExists(const aItem: Pointer): Boolean;
begin
  Result := Assigned(CheckNode(aItem));
end;

function TrTree.ItemExists(const aItem: Pointer): Boolean;
var
  i: Integer;
begin
  Result := IndexOf(aItem) >= 0;
  if not Result then
    for i := 0 to FNodes.Count - 1 do begin
      Result := NodeObjects[i].IndexOf(aItem) >= 0;
      if Result then Break;
    end;
end;

{$IFDEF POLARIS_D5}
procedure TrTree.SetOnNotify(const aValue: TrListNotifyEvent);
var
  i: Integer;
begin
  FOnNotify := aValue;
  for i := 0 to NodesCount - 1 do
    TrTree(FNodes.Items[i]).OnNotify := aValue;
end;
{$ENDIF}

function TrTree.GetNodeObject(const aIndex: Integer): TrTree;
begin
  Result := TrTree(FNodes.Items[aIndex]);
end;

procedure TrTree.DeleteNode(const aIndex: Integer);
begin
  FNodes.Delete(aIndex);
end;

procedure TrTree.ClearItems;
begin
  inherited Clear;
end;

procedure TrTree.ClearNodes;
begin
  if FNodes <> nil then FNodes.Clear;
end;

procedure TrTree.Clear;
begin
  ClearNodes;
  inherited;
end;

procedure TrTree.BeforeDestruction;
begin
  FNodes.Free;
  Fnodes := nil;
  inherited;
end;

{ TrLogFile }

procedure TrLogFile.AfterConstruction();
begin
  inherited;
  FCloseImmediately := True;
end;

procedure TrLogFile.SetCloseImmediately(const Value: Boolean);
begin
  FCloseImmediately := Value;
  if FCloseImmediately
    then Close();
end;

procedure TrLogFile.SetLogFileName(const Value: string);
begin
  FLogFileName := Value;
end;

procedure TrLogFile.WriteLog(const aText: string; AppendTime: Boolean = True);
var
  s: string;
begin
  if (nil = FStream) then begin
    if ('' = FLogFileName) then
      raise Exception.CreateFmt(srBadPropertyValue, ['LogFileName']);
    if FileExists(FLogFileName) then begin
      FStream := TFileStream.Create(FLogFileName, fmOpenWrite);
      FStream.Seek(0, soFromEnd);
    end
    else FStream := TFileStream.Create(FLogFileName, fmCreate);
  end;

  if AppendTime then begin
    s := DateTimeToStrFix(Now()) + ' ';
    FStream.Write(PChar(s)^, Length(s));
  end;
  FStream.Write(PChar(aText)^, Length(aText));
  FStream.Write(PChar(sLineBreak)^, Length(sLineBreak));

  if FCloseImmediately then
    Close()
{$ifdef mswindows}
   else FlushFileBuffers(FStream.Handle)
{$endif mswindows}
   ;
end;

procedure TrLogFile.Close();
begin
  if (nil <> FStream)
    then begin
    FStream.Free;
    FStream := nil;
  end;
end;

procedure TrLogFile.BeforeDestruction();
begin
  FStream.Free();
  inherited;
end;

{ TAppendString }

class function TAppendString.CreateNew: IAppendString;
begin
  Result := TAppendString.Create() as IAppendString;
end;

procedure TAppendString.BeginAppend();
begin
  FString := '';
  FLength := 0;
  FAppendFlag := True;
end;

procedure TAppendString.Append(const aString: string);
begin
  FAppendFlag := True;
  if Length(FString) < FLength + Length(aString) then
    SetLength(FString, Max(2*Length(FString), Length(aString)));
  Move(PChar(aString)^, PChar(Integer(PChar(FString)) + FLength)^, Length(aString));
  FLength := FLength + Length(aString);
end;

procedure TAppendString.EndAppend();
begin
  if FAppendFlag then begin
    SetLength(FString, FLength);
    FAppendFlag := False;
  end;
end;

function TAppendString.GetAsString(): string;
begin
  EndAppend();
  Result := FString;
end;

function TAppendString.GetCapacity: Integer;
begin
  Result := Length(FString);
end;

procedure TAppendString.SetCapacity(const Value: Integer);
begin
  if Value > Length(FString) then begin
    FAppendFlag := True;
    SetLength(FString, Value);
  end;
end;

end.

