(*====================================================================
    File rDLLloader.pas

    Components library for Delphi, Kylix and BCB++.
    Loader of DLL library.

    Copyright  2000-2008, PSTG
    http://www.polesoft.ru
    mailto:support@polesoft.ru

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

======================================================================
$Id: rDLLloader.pas,v 1.5 2008/09/05 09:00:45 savchenko Exp $
======================================================================*)
unit rDLLloader;

{$I POLARIS.INC}

interface

uses
  {$ifdef mswindows} Windows, {$endif}
  {$ifdef linux} Libc, {$endif}
  SysUtils, Classes;

type
  TrDLLloader = class(TComponent)
  private
    FActive: Boolean;
    FHInstance: HMODULE;
    FLibraryName: TFileName;
    FRequiredProcs: TStrings;
    FVersionInfo: TStrings;
    FAfterLoadLibrary: TNotifyEvent;
    FBeforeFreeLibrary: TNotifyEvent;
    FRequiredVersion: string;
    FFileVersion: string;
    {$ifdef linux}
    FLoadMode: Integer;
    {$endif}
    procedure ClearVersionInfo;
    function GetEntryPoint(const Name: string): Pointer;
    procedure InternalGetVersionInfo;
    procedure SetLibraryName(const Value: TFileName);
    procedure SetRequiredProcs(const Value: TStrings);
    procedure SetRequiredVersion(const Value: string);
  protected
    procedure CheckRequired;
    procedure DoAfterLoadLibrary; virtual;
    procedure DoBeforeFreeLibrary; virtual;
    procedure FreeDLL;
    property HInstance: HMODULE read FHInstance write FHInstance;
    procedure LoadDLL;
    procedure Loaded; override;
    procedure SetActive(const Value: Boolean); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Close;
    destructor Destroy; override;
    property EntryPoint[const Name: string]: Pointer read GetEntryPoint; default;
    procedure Open;
    property FileVersion: string read FFileVersion;
    {$ifdef linux}
    property LoadMode: Integer read FLoadMode write FLoadMode;
    {$endif}
  published
    property Active: Boolean read FActive write SetActive;
    property LibraryName: TFileName read FLibraryName write SetLibraryName;
    property RequiredProcs: TStrings read FRequiredProcs write SetRequiredProcs;
    property RequiredVersion: string read FRequiredVersion write SetRequiredVersion;
    property VersionInfo: TStrings read FVersionInfo stored False;
    property AfterLoadLibrary: TNotifyEvent read FAfterLoadLibrary write FAfterLoadLibrary;
    property BeforeFreeLibrary: TNotifyEvent read FBeforeFreeLibrary write FBeforeFreeLibrary;
  end;

  {$ifndef mswindows}
  DWORD = LongWord;
  {$endif}

implementation

uses
  rConst, rUtils;

{ TrDLLloader }

constructor TrDLLloader.Create(AOwner: TComponent);
begin
  inherited;
  FRequiredProcs := TStringList.Create;
  FVersionInfo := TStringList.Create;
  {$ifdef linux}
  FLoadMode := RTLD_LAZY;
  {$endif}
end;

destructor TrDLLloader.Destroy;
begin
  FreeDLL;
  FVersionInfo.Free;
  FRequiredProcs.Free;
  inherited;
end;

procedure TrDLLloader.CheckRequired;
var
  i: Integer;
  hProc: Pointer;
begin
  for i := 0 to FRequiredProcs.Count-1 do begin
    hProc := EntryPoint[FRequiredProcs[i]];
    if not Assigned(hProc) then
      raise Exception.CreateFmt(srDLLCheckRequiredError,[FRequiredProcs[i]]);
  end;
end;

procedure TrDLLloader.DoAfterLoadLibrary;
begin
  if Assigned(FAfterLoadLibrary) then
    FAfterLoadLibrary(Self);
end;

procedure TrDLLloader.DoBeforeFreeLibrary;
begin
  if Assigned(FBeforeFreeLibrary) then
    FBeforeFreeLibrary(Self);
end;

procedure TrDLLloader.FreeDLL;
begin
  if (FHInstance <> 0) then begin
    DoBeforeFreeLibrary;
    ClearVersionInfo;
    FreeLibrary(FHInstance);
    FHInstance := 0;
  end;
end;

procedure TrDLLloader.ClearVersionInfo;
begin
  FFileVersion := '';
  FVersionInfo.Clear;
end;

function TrDLLloader.GetEntryPoint(const Name: string): Pointer;
begin
  if FHInstance = 0 then
    Result := nil
  else
    Result := GetProcAddress(FHInstance, PChar(Name));
end;

procedure TrDLLloader.InternalGetVersionInfo;
begin
  {$ifdef mswindows}
  if FLibraryName = '' then Exit;
  FFileVersion := GetFileVersion(FLibraryName);
  GetVersionInfo(FVersionInfo, FLibraryName);
  {$endif}
  {$ifdef linux}
  FVersionInfo.Clear();
  {$endif}
end;

procedure TrDLLloader.LoadDLL;
begin
  InternalGetVersionInfo;
  if CompareVersion(FFileVersion, FRequiredVersion) = -1 then begin
    ClearVersionInfo;
    raise Exception.CreateFmt(srDLLCheckVersionError, [FLibraryName, FRequiredVersion]);
  end;
  {$ifdef mswindows}
{$IFDEF POLARIS_D6}
  FHInstance := SafeLoadLibrary(FLibraryName);
{$ELSE}
  FHInstance := LoadLibrary(PChar(FLibraryName));
{$ENDIF}
  {$endif}
  {$ifdef linux}
  FHInstance := HMODULE(dlopen(PChar(FLibraryName), FLoadMode));
  {$endif}
  if FHInstance = 0 then begin
    ClearVersionInfo;
    raise Exception.Create(Format(
      srDLLLoadError + sLineBreak + '%s', [FLibraryName, SysErrorMessage(GetLastError())]));
  end;
  try
    CheckRequired;
  except
    try
      FreeLibrary(FHInstance);
    finally
      FHInstance := 0;
    end;
    ClearVersionInfo;
    raise;
  end;
  DoAfterLoadLibrary;
end;

procedure TrDLLloader.Loaded;
begin
  inherited;
  if FActive then LoadDLL;
end;

procedure TrDLLloader.SetActive(const Value: Boolean);
begin
  if csReading in ComponentState
  then FActive := Value
  else if FActive <> Value then begin
    FActive := Value;
    FreeDLL;
    if FActive then
    try
      LoadDLL;
    except
      FActive := False;
      raise;
    end;
  end;
end;

procedure TrDLLloader.SetLibraryName(const Value: TFileName);
begin
  if csReading in ComponentState then FLibraryName := Value
  else if FLibraryName <> Value then begin
    FVersionInfo.Clear;
    if Active then raise Exception.Create(srDLLChangeNameError);
    FLibraryName := Value;
    {$ifdef mswindows}
    if FLibraryName <> '' then GetVersionInfo(FVersionInfo, FLibraryName);
    {$endif}
  end;
end;

procedure TrDLLloader.SetRequiredProcs(const Value: TStrings);
begin
  FRequiredProcs.Assign(Value);
end;

procedure TrDLLloader.SetRequiredVersion(const Value: string);
var
  MS, LS: DWORD;
begin
  if SeparateVersion(Value, MS, LS) then
    FRequiredVersion := Value
  else
    raise Exception.Create(srDLLVersionFormatError);
end;

procedure TrDLLloader.Close;
begin
  Active := False;
end;

procedure TrDLLloader.Open;
begin
  Active := True;
end;

end.

