(*====================================================================
    File rPlugins.pas

    Components library for Delphi and BCB++.
      .
    Plugins support.

    Copyright  1998-2008, PSTG
    http://www.polesoft.ru
    mailto:support@polesoft.ru

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

======================================================================
$Id: rPlugins.pas,v 1.3 2008/09/05 09:00:45 savchenko Exp $
======================================================================*)
unit rPlugins;

{$I POLARIS.INC}

interface

uses
  {$ifdef mswindows} Windows, Registry, {$endif}
  Classes, SysUtils, rUtils, rConst,
  rDLLloader, {$IFDEF POLARIS_D6}RxStrUtils{$ELSE}StrUtils{$ENDIF};

//=============================================================================

type
  TrPluginLoadError = procedure(Sender: TObject; PluginLibrary, ErrorMessage: string) of object;

  TrPlugin = class(TrDLLloader)
  private
    fPlugName: string;
    fPlugType: string;
    fDefault: Boolean;
    fOptParams: TStringList;
    fInstCount: Integer;
    procedure UpdateInfo;
    procedure SetDefault(Value: Boolean);
    function CheckRequiredSoft: Boolean;
    function GetOptParams: string;
    function GetOptParam(aParamName: string): string;
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
    function Execute(const Params1, Params2, Params3: string): string;
    procedure IncInstCount;
    procedure DecInstCount;
    property PlugName: string read FPlugName;
    property PlugType: string read FPlugType;
    property OptParams: string read GetOptParams;
    property OptParam[aParamName: string]: string read GetOptParam;
    property IsDefault: Boolean read FDefault;
    property InstCount: Integer read fInstCount;
  end;

  TrPlugins = class(TComponent)
  private
    fPlugins: TList;
    fRegKey: string;
    fPlugsPath: string;
    fFileMask: string;
    fRequiredVersion: string;
    {$ifdef mswindows}
    fReg: TRegistry;
    {$endif}
    FOnPluginLoadError: TrPluginLoadError;
    function GetPlugin(Number: Integer): TrPlugin;
    procedure SetPlugsPath(const Value: string);
    function CheckVersionFormat: Boolean;
    function GetCount: Integer;
    procedure SetRequiredVersion(const Value: string);
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
    procedure SaveParamsToReg;
    property Plugins[Index: Integer]: TrPlugin read GetPlugin;
    function ExecutePlugin(
      const PlugName: string;
      const PlugType: string;
      const Param1, Param2, Param3: string): string;
    function GetPlugTypes: string;
    function GetPlugNames(const PlugType: string): string;
    function GetDefaultPlug(const PlugType: string): string;
    function GetPlugByOptParams(PlugType, ParamName, ParamValue: string): string;
    function GetExistsPlug(
      const PlugType, PlugName: string;
      DefaultIfNotExists: Boolean {$IFDEF POLARIS_D4}= False{$ENDIF}
      ): string;
    function GetPluginByName(const PlugName: string): TrPlugin;
    function GetDefaultPlugin(const PlugType: string): TrPlugin;
    function PlugExists(const PlugType, PlugName: string): Boolean;
    procedure SetDefaultPlug(const PlugType, PlugName: string);
    property PlugsPath: string read FPlugsPath write SetPlugsPath;
    property Count: Integer read GetCount;
  published
    property RegKey: string read FRegKey write FRegKey;
    property FileMask: string read FFileMask write FFileMask stored True;
    property RequiredVersion: string read FRequiredVersion write SetRequiredVersion
      stored CheckVersionFormat;
    property OnPluginLoadError: TrPluginLoadError read FOnPluginLoadError write FOnPluginLoadError;
  end;

implementation

const
  cFileMask =
    {$ifdef mswindows} '*.dll'; {$endif}
    {$ifdef linux}     '*.so';  {$endif}
  cExecuteProc = 'Execute';
  cGetInfoProc = 'GetInfo';

//=============================================================================
//
// = TrPlugin =
//
//=============================================================================

constructor TrPlugin.Create(aOwner: TComponent);
begin
  inherited;
  fOptParams := TStringList.Create;
  fOptParams.Sorted := True;
  fOptParams.Duplicates := dupIgnore;
  fDefault := False;
  fInstCount := 0;
end;

destructor TrPlugin.Destroy;
begin
  fOptParams.Free;
  inherited;
end;

procedure TrPlugin.SetDefault(Value: Boolean);
begin
  fDefault := Value;
end;

function TrPlugin.CheckRequiredSoft: Boolean;
begin
  if not Active
    then Result := False
    else Result := Assigned(EntryPoint[cExecuteProc]) and
                   Assigned(EntryPoint[cGetInfoProc]);
end;

function TrPlugin.Execute(const Params1, Params2, Params3: string): string;
type
  TExecuteProc = function(const Params1, Params2, Params3: pChar): pChar; stdcall;
var
  vNotActive: Boolean;
begin
  vNotActive := not Active;
  if vNotActive then Open;
  Result := TExecuteProc(EntryPoint[cExecuteProc])
    (pChar(Params1), pChar(Params2), pChar(Params3));
  if vNotActive then Close;
end;

procedure TrPlugin.UpdateInfo;
type
  TInfoProc = procedure(
    var sFullName: pChar;
    var sPlugType: pChar;
    var sOptParams: pChar); stdcall;
var
  p1, p2, p3: pChar;
begin
  TInfoProc(EntryPoint[cGetInfoProc])(p1, p2, p3);
  fPlugName := p1;
  fPlugType := p2;
  fOptParams.Text := p3;
end;

function TrPlugin.GetOptParams: string;
begin
  Result := FOptParams.Text;
end;

function TrPlugin.GetOptParam(aParamName: string): string;
begin
  Result := FOptParams.Values[aParamName];
end;

procedure TrPlugin.IncInstCount;
begin
  Inc(fInstCount);
end;

procedure TrPlugin.DecInstCount;
begin
  Dec(fInstCount);
  if fInstCount < 1 then begin
    Close;
    fInstCount := 0;
  end;
end;

//=============================================================================
//
// = TrPlugins =
//
//=============================================================================
constructor TrPlugins.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  {$ifdef mswindows}
  fReg := TRegistry.Create;
  {$endif}
  fPlugins := TList.Create;
  fFileMask := cFileMask;
end;

//=============================================================================
destructor TrPlugins.Destroy;
begin
  SaveParamsToReg;
  fPlugins.Free;
  {$ifdef mswindows}
  fReg.Free;
  {$endif}
  inherited Destroy;
end;

//=============================================================================
//    
function TrPlugins.GetPlugin(Number: Integer): TrPlugin;
begin
  Result := TrPlugin(FPlugins[Number]);
end;

//=============================================================================
//    
function TrPlugins.GetPluginByName(const PlugName: string): TrPlugin;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to FPlugins.Count - 1 do
    if Plugins[i].PlugName = PlugName then begin
      Result := TrPlugin(FPlugins[i]);
      break;
    end;
end;

//=============================================================================
//   
function TrPlugins.GetDefaultPlugin(const PlugType: string): TrPlugin;
var
  i: Integer;
begin
  Result := nil;
  if PlugType <> '' then
    for i := 0 to FPlugins.Count - 1 do
      if (Plugins[i].PlugType = PlugType) and (Plugins[i].IsDefault) then begin
        Result := Plugins[i];
        Break;
      end;
end;

//=============================================================================
//  
function TrPlugins.ExecutePlugin(
  const PlugName: string;
  const PlugType: string;
  const Param1, Param2, Param3 : string): string;
var
  i : Integer;
begin
  Result := Format(srPluginNotFound, [PlugName, PlugType]);
  for i := 0 to FPlugins.Count - 1 do
    if ( (PlugName = Plugins[i].PlugName) and
         ( (PlugType = '') or (PlugType = Plugins[i].PlugType) )
       )
       or
       ( (PlugType = Plugins[i].PlugType) and (PlugName = '') and
         (Plugins[i].IsDefault)
       )
    then begin
      Result := Plugins[i].Execute(Param1, Param2, Param3);
      break;
    end;
end;

//=============================================================================
//   
procedure TrPlugins.SetPlugsPath(const Value: string);
var
  i: Integer;
  sr: TSearchRec;
  sl: TStringList;
  {$ifdef mswindows}
  r: HKEY;
  s: string;
  {$endif}
  sError: string;
  plug: TrPlugin;
  vbOpenError: Boolean;
begin

  plug := nil;
  sl := TStringList.Create;
  try

    for i := fPlugins.Count - 1 downto 0 do begin
      Plugins[i].Free;
      FPlugins.Delete(i);
    end;

    fPlugsPath := Value;
    if FindFirst(AddSlash(FPlugsPath) + FFileMask, faAnyFile, sr) = 0 then begin
      sl.Add(sr.Name);
      while FindNext(sr) = 0 do sl.Add(sr.Name);
    end;
    FindClose(sr);

    sError := '';
    try
      for i := 0 to sl.Count - 1 do begin
        if plug = nil then begin
          plug := TrPlugin.Create(Self);
          if FRequiredVersion <> '' then plug.RequiredVersion := FRequiredVersion;
        end;
        plug.LibraryName := AddSlash(FPlugsPath) + sl.Strings[i];
        vbOpenError := False;
        try
          plug.Open;
        except
          on E: Exception do begin
            sError := sError + crlf + E.Message;
            vbOpenError := True;
          end;
        end;
        if vbOpenError and Assigned(FOnPluginLoadError) then begin
          FOnPluginLoadError(Self, plug.LibraryName, sError);
          sError := '';
        end;
        if not vbOpenError and plug.CheckRequiredSoft then begin
          plug.UpdateInfo;
          if not PlugExists('', plug.PlugName) then begin
            FPlugins.Add(plug);
            plug := nil;
          end;
        end;
        if plug <> nil then plug.Close;
      end;
      if plug <> nil then begin
        plug.Free;
        plug := nil;
      end;
    finally
      for i := 0 to fPlugins.Count - 1 do Plugins[i].Close;
    end;

    sl.Text := GetPlugTypes;
    for i := 0 to sl.Count - 1 do
      SetDefaultPlug(sl.Strings[i], ExtractWord(1, GetPlugNames(sl.Strings[i]), [#13]));
    {$ifdef mswindows}
    if FRegKey <> '' then begin
      SeparateRegKey(FRegKey, r, s);
      fReg.RootKey := r;
      fReg.OpenKey(s + '\Plugins\Default', True);
      for i := 0 to sl.Count - 1 do
        if fReg.ValueExists(sl.Strings[i])
          then SetDefaultPlug(sl.Strings[i], fReg.ReadString(sl.Strings[i]));
    end;
    {$endif}

    if sError <> '' then raise Exception.Create(sError);

  finally
    {$ifdef mswindows}
    fReg.CloseKey;
    {$endif}
    sl.Free;
    if plug <> nil then plug.Free;
  end;

end;

//=============================================================================
//   
function TrPlugins.GetPlugTypes: string;
var
  i  : Integer;
  sl : TStringList;
begin
  sl := TStringList.Create;
  sl.Sorted := True;
  sl.Duplicates := dupIgnore;
  for i := 0 to FPlugins.Count - 1 do sl.Add(Plugins[i].PlugType);
  Result := sl.Text;
  sl.Free;
end;

//=============================================================================
//  
function TrPlugins.GetPlugNames(const PlugType: string): string;
var
  i  : Integer;
  sl : TStringList;
begin
  sl := TStringList.Create;
  sl.Sorted := True;
  sl.Duplicates := dupIgnore;
  for i := 0 to FPlugins.Count - 1 do
    if (PlugType = '') or (PlugType = Plugins[i].PlugType)
      then sl.Add(Plugins[i].PlugName);
  Result := sl.Text;
  sl.Free;
end;

//=============================================================================
//    
procedure TrPlugins.SetDefaultPlug(const PlugType, PlugName: string);
var
  i: Integer;
begin
  if PlugExists(PlugType, PlugName) and (PlugName <> '') then
    for i := 0 to FPlugins.Count - 1 do
      if Plugins[i].PlugType = PlugType then
        Plugins[i].SetDefault(Plugins[i].PlugName = PlugName);
end;

//=============================================================================
//    
function TrPlugins.GetDefaultPlug(const PlugType: string): string;
var
  p: TrPlugin;
begin
  p := GetDefaultPlugin(PlugType);
  if Assigned(p) then Result := p.PlugName else Result := '';
end;

//=============================================================================
//   
function TrPlugins.PlugExists(const PlugType, PlugName: string): Boolean;
var
  r : Integer;
  sl  : TStringList;
begin
  sl := TStringList.Create;
  sl.Sorted := True;
  sl.Text := GetPlugNames(PlugType);
  if sl.Count > 0 then begin
    if PlugName = ''
      then Result := True
      else Result := sl.Find(PlugName, r);
  end
  else Result := False;
  sl.Free;
end;

//=============================================================================
//      
function TrPlugins.GetPlugByOptParams(PlugType, ParamName, ParamValue: string
  ): string;
var
  i: Integer;
  p: TrPlugin;
begin
  Result := '';
  p := GetDefaultPlugin(PlugType);
  if Assigned(p) and (p.OptParam[ParamName] = ParamValue)
    then Result := p.PlugName
    else
      for i := 0 to FPlugins.Count - 1 do
        if ((PlugType = '') or (Plugins[i].PlugType = PlugType)) and
           (Plugins[i].OptParam[ParamName] = ParamValue)
        then begin
          Result := Plugins[i].PlugName;
          Break;
        end;
end;

//=============================================================================
//    ( )     ( )
function TrPlugins.GetExistsPlug(
  const PlugType, PlugName: string;
  DefaultIfNotExists: Boolean
  ): string;
begin
  Result := '';
  if (PlugType <> '') and DefaultIfNotExists and
    ( (PlugName = '') or not PlugExists('', PlugName) )
    then Result := GetDefaultPlug(PlugType)
  else
  if (PlugName <> '') and PlugExists(PlugType, PlugName) then Result := PlugName;
end;

//=============================================================================
//   
function TrPlugins.CheckVersionFormat: Boolean;
var
  MS, LS: DWORD;
begin
  Result := SeparateVersion(fRequiredVersion, MS, LS);
end;

//=============================================================================
//      
procedure TrPlugins.SetRequiredVersion(const Value: string);
var
  MS, LS: DWORD;
begin
  if SeparateVersion(Value, MS, LS) then fRequiredVersion := Value;
end;

//=============================================================================
//  
procedure TrPlugins.SaveParamsToReg;
{$ifdef mswindows}
var
  sl: TStringList;
  i: Integer;
  s: string;
  r: HKEY;
{$endif}
begin
  {$ifdef mswindows}
  if fRegKey <> '' then begin
    sl := TStringList.Create;
    try
      sl.Text := GetPlugTypes;
      SeparateRegKey(fRegKey, r, s);
      fReg.RootKey := r;
      fReg.OpenKey(s + '\Plugins\Default', True);
      for i := 0 to sl.Count - 1 do
        fReg.WriteString(sl.Strings[i], GetDefaultPlug(sl.Strings[i]));
    finally
      fReg.CloseKey;
      sl.Free;
    end;
  end;
  {$endif}
end;

//=============================================================================
function TrPlugins.GetCount: Integer;
begin
  Result := fPlugins.Count;
end;

//=============================================================================
end.

